/*
 * Copyright Camunda Services GmbH and/or licensed to Camunda Services GmbH under
 * one or more contributor license agreements. See the NOTICE file distributed
 * with this work for additional information regarding copyright ownership.
 * Licensed under the Camunda License 1.0. You may not use this file
 * except in compliance with the Camunda License 1.0.
 */
package io.camunda.optimize.service.db.os.report.service;

import static io.camunda.optimize.service.db.os.externalcode.client.dsl.QueryDSL.filter;
import static io.camunda.optimize.service.db.os.report.interpreter.util.FilterLimitedAggregationUtilOS.wrapWithFilterLimitedParentAggregation;
import static io.camunda.optimize.service.db.os.report.interpreter.util.NumberHistogramAggregationUtilOS.generateHistogramFromScript;
import static io.camunda.optimize.service.db.schema.index.ProcessInstanceIndex.DURATION;
import static io.camunda.optimize.service.db.schema.index.ProcessInstanceIndex.START_DATE;

import io.camunda.optimize.dto.optimize.query.report.single.configuration.SingleReportConfigurationDto;
import io.camunda.optimize.dto.optimize.query.report.single.configuration.custom_buckets.CustomBucketDto;
import io.camunda.optimize.dto.optimize.query.report.single.filter.data.operator.ComparisonOperator;
import io.camunda.optimize.dto.optimize.query.report.single.process.ProcessReportDataDto;
import io.camunda.optimize.dto.optimize.query.report.single.process.filter.data.DurationFilterDataDto;
import io.camunda.optimize.service.db.os.report.interpreter.RawResult;
import io.camunda.optimize.service.db.os.report.interpreter.distributedby.process.ProcessDistributedByInterpreterFacadeOS;
import io.camunda.optimize.service.db.os.report.interpreter.util.DurationScriptUtilOS;
import io.camunda.optimize.service.db.os.report.interpreter.util.FilterLimitedAggregationUtilOS;
import io.camunda.optimize.service.db.report.ExecutionContext;
import io.camunda.optimize.service.db.report.MinMaxStatDto;
import io.camunda.optimize.service.db.report.plan.process.ProcessExecutionPlan;
import io.camunda.optimize.service.db.report.result.CompositeCommandResult;
import io.camunda.optimize.service.db.report.result.CompositeCommandResult.DistributedByResult;
import io.camunda.optimize.service.db.report.service.DurationAggregationService;
import io.camunda.optimize.service.security.util.LocalDateUtil;
import io.camunda.optimize.service.util.configuration.condition.OpenSearchCondition;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.BiFunction;
import lombok.AllArgsConstructor;
import org.apache.commons.lang3.tuple.Pair;
import org.opensearch.client.opensearch._types.Script;
import org.opensearch.client.opensearch._types.aggregations.Aggregate;
import org.opensearch.client.opensearch._types.aggregations.Aggregation;
import org.opensearch.client.opensearch._types.query_dsl.Query;
import org.opensearch.client.opensearch.core.SearchResponse;
import org.springframework.context.annotation.Conditional;
import org.springframework.stereotype.Component;

@Component
@AllArgsConstructor
@Conditional(OpenSearchCondition.class)
public class DurationAggregationServiceOS extends DurationAggregationService {
  private final MinMaxStatsServiceOS minMaxStatsService;
  private final ProcessDistributedByInterpreterFacadeOS distributedByInterpreter;

  public Optional<Pair<String, Aggregation>> createLimitedGroupByScriptedDurationAggregation(
      final Query baseQuery,
      final ExecutionContext<ProcessReportDataDto, ProcessExecutionPlan> context,
      final Script durationCalculationScript) {
    final MinMaxStatDto minMaxStats =
        minMaxStatsService.getMinMaxNumberRangeForScriptedField(
            context, baseQuery, getIndexNames(context), durationCalculationScript);
    return createLimitedGroupByScriptedDurationAggregation(
        context,
        baseQuery,
        durationCalculationScript,
        minMaxStats,
        this::createProcessInstanceLimitingFilterQuery);
  }

  public List<CompositeCommandResult.GroupByResult> mapGroupByDurationResults(
      final SearchResponse<RawResult> response,
      final Map<String, Aggregate> parentSubAggregations,
      final ExecutionContext<ProcessReportDataDto, ProcessExecutionPlan> context) {
    return Optional.ofNullable(parentSubAggregations)
        .flatMap(FilterLimitedAggregationUtilOS::unwrapFilterLimitedAggregations)
        .map(aggregations -> aggregations.get(DURATION_HISTOGRAM_AGGREGATION).histogram())
        .stream()
        .flatMap(
            histogramAggregation ->
                histogramAggregation.buckets().array().stream()
                    .map(
                        durationBucket -> {
                          final List<DistributedByResult> distributions =
                              distributedByInterpreter.retrieveResult(
                                  response, durationBucket.aggregations(), context);
                          return CompositeCommandResult.GroupByResult.createGroupByResult(
                              String.valueOf(durationBucket.key()), distributions);
                        }))
        .toList();
  }

  private Optional<Pair<String, Aggregation>> createLimitedGroupByScriptedDurationAggregation(
      final ExecutionContext<ProcessReportDataDto, ProcessExecutionPlan> context,
      final Query baseQuery,
      final Script durationCalculationScript,
      final MinMaxStatDto minMaxStats,
      final BiFunction<ComparisonOperator, Double, Query> limitingFilterCreator) {

    if (minMaxStats.isEmpty()) {
      return Optional.empty();
    }

    final SingleReportConfigurationDto reportConfigurationDto =
        context.getReportData().getConfiguration();

    final CustomBucketDto customBucketDto = reportConfigurationDto.getCustomBucket();
    final double minValueInMillis = getMinValueInMillis(minMaxStats, customBucketDto);
    final double maxValueInMillis = minMaxStats.getMax();
    if (minValueInMillis > maxValueInMillis) {
      return Optional.empty();
    }

    final double intervalInMillis =
        getIntervalInMillis(minValueInMillis, maxValueInMillis, customBucketDto);

    final Query limitingFilter =
        filter(
            limitingFilterCreator.apply(ComparisonOperator.GREATER_THAN_EQUALS, minValueInMillis));

    final Pair<String, Aggregation> histogramAggregation =
        generateHistogramFromScript(
            DURATION_HISTOGRAM_AGGREGATION,
            intervalInMillis,
            minValueInMillis,
            durationCalculationScript,
            maxValueInMillis,
            distributedByInterpreter.createAggregations(context, baseQuery));

    return Optional.of(
        wrapWithFilterLimitedParentAggregation(limitingFilter, histogramAggregation));
  }

  private Query createProcessInstanceLimitingFilterQuery(
      final ComparisonOperator filterOperator, final double filterValueInMillis) {
    return createLimitingFilterQuery(
        filterOperator, (long) filterValueInMillis, DURATION, START_DATE, false);
  }

  private Query createLimitingFilterQuery(
      final ComparisonOperator filterOperator,
      final long filterValueInMillis,
      final String durationFieldName,
      final String referenceDateFieldName,
      final boolean includeNull) {
    return new Query.Builder()
        .script(
            b ->
                b.script(
                    DurationScriptUtilOS.getDurationFilterScript(
                        LocalDateUtil.getCurrentDateTime().toInstant().toEpochMilli(),
                        durationFieldName,
                        referenceDateFieldName,
                        DurationFilterDataDto.builder()
                            .operator(filterOperator)
                            .unit(FILTER_UNIT)
                            .value(filterValueInMillis)
                            .includeNull(includeNull)
                            .build())))
        .build();
  }
}
