/*
 * Copyright Camunda Services GmbH and/or licensed to Camunda Services GmbH under
 * one or more contributor license agreements. See the NOTICE file distributed
 * with this work for additional information regarding copyright ownership.
 * Licensed under the Camunda License 1.0. You may not use this file
 * except in compliance with the Camunda License 1.0.
 */
package io.camunda.optimize.service.db.schema;

import io.camunda.optimize.service.db.DatabaseClient;
import io.camunda.optimize.service.util.configuration.ConfigurationService;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Component;

@Component
@Slf4j
@Setter
public abstract class DatabaseSchemaManager<CLIENT extends DatabaseClient, BUILDER> {

  protected ConfigurationService configurationService;
  protected OptimizeIndexNameService indexNameService;

  @Getter protected final List<IndexMappingCreator<BUILDER>> mappings;

  protected DatabaseSchemaManager(
      final ConfigurationService configurationService,
      final OptimizeIndexNameService indexNameService,
      final List<IndexMappingCreator<BUILDER>> mappings) {
    this.configurationService = configurationService;
    this.indexNameService = indexNameService;
    this.mappings = mappings;
  }

  public abstract void validateDatabaseMetadata(final CLIENT dbClient);

  public abstract void initializeSchema(final CLIENT dbClient);

  public abstract boolean schemaExists(CLIENT dbClient);

  public abstract boolean indexExists(
      final CLIENT dbClient, final IndexMappingCreator<BUILDER> mapping);

  public abstract boolean indexExists(final CLIENT dbClient, final String indexName);

  public abstract boolean indicesExist(
      final CLIENT dbClient, final List<IndexMappingCreator<BUILDER>> mappings);

  public abstract void createIndexIfMissing(
      final CLIENT dbClient, final IndexMappingCreator<BUILDER> indexMapping);

  public abstract void createIndexIfMissing(
      final CLIENT dbClient,
      final IndexMappingCreator<BUILDER> indexMapping,
      final Set<String> additionalReadOnlyAliases);

  public abstract void createOrUpdateOptimizeIndex(
      final CLIENT dbClient,
      final IndexMappingCreator<BUILDER> mapping,
      final Set<String> readOnlyAliases);

  public abstract void deleteOptimizeIndex(
      final CLIENT dbClient, final IndexMappingCreator<BUILDER> mapping);

  public abstract void createOrUpdateTemplateWithoutAliases(
      final CLIENT dbClient, final IndexMappingCreator<BUILDER> mappingCreator);

  public abstract void updateDynamicSettingsAndMappings(
      CLIENT dbClient, IndexMappingCreator<BUILDER> indexMapping);

  public void addMapping(final IndexMappingCreator<BUILDER> mapping) {
    mappings.add(mapping);
  }

  public void createOptimizeIndices(CLIENT dbClient) {
    for (IndexMappingCreator<BUILDER> mapping : mappings) {
      createOrUpdateOptimizeIndex(dbClient, mapping);
    }
  }

  public void createOrUpdateOptimizeIndex(
      final CLIENT dbClient, final IndexMappingCreator<BUILDER> mapping) {
    createOrUpdateOptimizeIndex(dbClient, mapping, Collections.emptySet());
  }
}
