/*
 * Copyright Camunda Services GmbH and/or licensed to Camunda Services GmbH under
 * one or more contributor license agreements. See the NOTICE file distributed
 * with this work for additional information regarding copyright ownership.
 * Licensed under the Camunda License 1.0. You may not use this file
 * except in compliance with the Camunda License 1.0.
 */
package io.camunda.tasklist.it;

import static org.assertj.core.api.Assertions.assertThat;

import io.camunda.tasklist.property.TasklistProperties;
import io.camunda.tasklist.property.ZeebeProperties;
import io.camunda.tasklist.qa.util.ContainerVersionsUtil;
import io.camunda.tasklist.util.CertificateUtil;
import io.camunda.tasklist.zeebe.ZeebeConnector;
import io.camunda.zeebe.client.ZeebeClient;
import io.camunda.zeebe.client.api.response.BrokerInfo;
import io.zeebe.containers.ZeebeContainer;
import java.io.File;
import java.time.Duration;
import java.util.List;
import java.util.Map;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.api.io.TempDir;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.testcontainers.containers.wait.strategy.LogMessageWaitStrategy;
import org.testcontainers.utility.DockerImageName;
import org.testcontainers.utility.MountableFile;

@SpringBootTest(
    classes = {ZeebeConnector.class, TasklistProperties.class},
    properties = {
      TasklistProperties.PREFIX + ".importer.startLoadingDataOnStartup = false",
      TasklistProperties.PREFIX + ".archiver.rolloverEnabled = false",
    })
@ExtendWith(SpringExtension.class)
public class ZeebeConnectorSecureIT {

  private static final String CERTIFICATE_FILE = "zeebe-test-chain.cert.pem";
  private static final String PRIVATE_KEY_FILE = "zeebe-test-server.key.pem";
  private static final DockerImageName ZEEBE_DOCKER_IMAGE =
      DockerImageName.parse("camunda/zeebe")
          .withTag(
              ContainerVersionsUtil.readProperty(
                  ContainerVersionsUtil.ZEEBE_CURRENTVERSION_DOCKER_PROPERTY_NAME));
  @Autowired ZeebeConnector zeebeConnector;
  private ZeebeContainer zeebeContainer;
  private ZeebeClient zeebeClient;

  @Test
  public void shouldConnectWithTLS(@TempDir final File tempDir) throws Exception {
    // given
    final File certFile = new File(tempDir, CERTIFICATE_FILE);
    final File privateKeyFile = new File(tempDir, PRIVATE_KEY_FILE);
    CertificateUtil.generateRSACertificate(certFile, privateKeyFile);
    zeebeContainer =
        new ZeebeContainer(ZEEBE_DOCKER_IMAGE)
            .withCopyFileToContainer(
                MountableFile.forHostPath(tempDir.toPath(), 0755), "/usr/local/zeebe/certs")
            .withEnv(
                Map.of(
                    "ZEEBE_BROKER_GATEWAY_SECURITY_CERTIFICATECHAINPATH",
                    "/usr/local/zeebe/certs/" + CERTIFICATE_FILE,
                    "ZEEBE_BROKER_GATEWAY_SECURITY_PRIVATEKEYPATH",
                    "/usr/local/zeebe/certs/" + PRIVATE_KEY_FILE,
                    "ZEEBE_BROKER_GATEWAY_SECURITY_ENABLED",
                    "true"))
            // Can't use connection wait strategy because of TLS
            .waitingFor(
                new LogMessageWaitStrategy()
                    .withRegEx(".*Broker is ready!.*")
                    .withStartupTimeout(Duration.ofSeconds(101)));
    zeebeContainer.start();
    zeebeClient =
        zeebeConnector.newZeebeClient(
            new ZeebeProperties()
                .setGatewayAddress(zeebeContainer.getExternalGatewayAddress())
                .setSecure(true)
                .setCertificatePath(tempDir.getCanonicalPath() + "/" + CERTIFICATE_FILE));
    // when
    final List<BrokerInfo> brokerInfos =
        zeebeClient.newTopologyRequest().send().join().getBrokers();
    // then
    assertThat(brokerInfos).isNotEmpty();
  }

  @AfterEach
  public void cleanUp() {
    if (zeebeClient != null) {
      zeebeClient.close();
    }
    if (zeebeContainer != null) {
      zeebeContainer.stop();
    }
  }
}
